from struct import unpack, pack

from elf import valid_address, get_offset
from php5.entry import get_glob_segments, get_data
from php5.my_types import class_entry_size_partial, string_deprecation_size, hashtable_size, spacer

name = 8
name_len = name + 8
parent = name_len + 8
function_table = parent + 16
properties_info = function_table + hashtable_size
default_properties_table_pp = properties_info + hashtable_size
default_static_members_table_pp = default_properties_table_pp + 8
static_members_table_pp = default_static_members_table_pp + 8
constants_table = static_members_table_pp + 8
default_properties_count = constants_table + hashtable_size
default_static_members_count = default_properties_count + 4


class ZendClassEntry:
    def __init__(self, address, level=0) -> None:
        self.address = address
        self.level = level
        self.valid_address = valid_address(self.address, get_glob_segments())
        self.valid = False
        self.name = ''
        if self.valid_address:
            self.offset = get_offset(self.address, get_glob_segments())
            self.data = get_data(self.offset, class_entry_size_partial)
            (self.name_p, self.name_len, pad, self.parent), \
            (self.default_properties_table_pp, self.default_static_members_table_pp, self.static_members_table_pp), \
            (self.default_properties_count, self.default_static_members_count) = self.parse_class_entry()
            for i in [1]:
                self.valid_name = self.check_name()
                if not self.valid_name:
                    continue
                self.name = self.get_name()

                if not self.valid_pointer(self.parent):
                    continue

                from php5.my_hashtable import ArrayZvalue
                self.function_table = ArrayZvalue(pack('QQ', self.address + function_table, 0), self.level + 2)
                if not (self.function_table.is_valid() or self.function_table.is_empty()):
                    continue

                self.properties_info = ArrayZvalue(pack('QQ', self.address + properties_info, 0), self.level + 2)
                if not (self.function_table.is_valid() or self.function_table.is_empty()):
                    continue

                self.constants_table = ArrayZvalue(pack('QQ', self.address + constants_table, 0), self.level + 2)
                if not (self.constants_table.is_valid() or self.constants_table.is_empty()):
                    continue

                self.default_properties_table_p = self.get_pointer_at_address(address + default_properties_table_pp)
                self.default_static_members_table_p = self.get_pointer_at_address(
                    address + default_static_members_table_pp)
                self.static_members_table_p = self.get_pointer_at_address(address + static_members_table_pp)

                if not self.valid_pointer_len(self.default_properties_table_p, self.default_properties_count):
                    continue

                if not self.valid_pointer_len(self.default_static_members_table_p, self.default_static_members_count):
                    continue

                if not self.valid_pointer(self.static_members_table_p):
                    continue

                if self.default_properties_table_p:
                    self.default_properties_table = self.get_zval_array(self.default_properties_table_p,
                                                                        self.default_properties_count)
                else:
                    self.default_properties_table = []

                if self.default_static_members_table_p:
                    self.default_static_members_table = self.get_zval_array(self.default_static_members_table_p,
                                                                            self.default_static_members_count)
                else:
                    self.default_static_members_table = []

                if self.static_members_table_p:
                    self.static_members_table = self.get_zval_array(self.static_members_table_p, 1)
                else:
                    self.static_members_table = []

                self.valid = True

    def is_valid(self):
        return self.valid

    def get_name(self):
        if valid_address(self.name_p, get_glob_segments()):
            return get_data(get_offset(self.name_p, get_glob_segments()), self.name_len)
        return ''

    def parse_class_entry(self):
        return unpack('QIiQ', self.data[8:32]), unpack('QQQ', self.data[-104:-80]), unpack('ii', self.data[-8:])

    def check_name(self):
        return (self.name_p == self.name_len == 0) or (
            valid_address(self.name_p, get_glob_segments()) and 0 <= self.name_len < string_deprecation_size)

    def valid_pointer(self, address):
        return valid_address(address, get_glob_segments()) or address == 0

    def valid_pointer_len(self, address, len):
        return valid_address(address, get_glob_segments()) or address == len == 0

    def get_pointer_at_address(self, address):
        offset = get_offset(address, get_glob_segments())
        return unpack('Q', get_data(offset, 8))[0]

    def get_zval_array(self, address, count):
        result = []
        for index in range(count):
            pointer = self.get_pointer_at_address(address + 8 * index)
            from php5.my_zval_struct import ZvalStruct
            zval = ZvalStruct(pointer, self.level + 2)
            result.append(zval)
        return result

    def get_properties_info_size(self):
        return self.properties_info.get_bucket_list_size()

    def get_data_to_print(self):
        indent1 = self.level * spacer
        indent2 = (self.level + 1) * spacer
        if not self.is_valid():
            return '{}Invalid Class Entry\n'.format(indent1)
        args = {'indent1': indent1, 'indent2': indent2, 'name': self.name, 'ft': self.function_table,
                'pi': self.properties_info, 'ct': self.constants_table,
                'dpt': '\n'.join([str(x) for x in self.default_properties_table]),
                'dsmt': '\n'.join([str(x) for x in self.default_static_members_table]),
                'smt': '\n'.join([str(x) for x in self.static_members_table])}
        return '{indent1}Class entry:\n{indent2}Name: {name}\n{indent2}Function table:\n{ft}\n{indent2}Properties info:\n{pi}\n{indent2}Default properties table:\n{dpt}\n{indent2}Default static members table:\n{dsmt}\n{indent2}Static members table:\n{smt}\n{indent2}Constants table:\n{ct}\n'.format(
            **args)

    def __repr__(self) -> str:
        return self.get_data_to_print()

    def __str__(self) -> str:
        return self.get_data_to_print()


# 0x7f3fcc5456a8
# 0x7f3fcb7b65d0

if __name__ == "__main__":
    filename = '../../dumps/php5/core.14197'
    from php5.entry import init

    init(filename)
    ce1 = ZendClassEntry(0x7f3fcc5456a8)
    ce2 = ZendClassEntry(0x7f3fcb7b65d0)
    # print(ce2)

    # 0x7f3fcc545690
    from php5.my_bucket import Bucket

    bucket = Bucket(0x7f3fcc545690, special_type=ZendClassEntry)
    print(bucket)

    from elf import get_all_addresses

    for addr in get_all_addresses(get_glob_segments()):
        ze = ZendClassEntry(addr)
        if ze.is_valid():
            print(ze)
